package com.dsk.flink.dsc.common.function;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.map.MapUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.utils.EnvProperties;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.async.ResultFuture;
import org.apache.flink.streaming.api.functions.async.RichAsyncFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author shezaixing
 * @date 2023/12/6 15:40
 * @description mysql/tidb sink 数据组装 sql
 *
 */
public class AsyncMysqlDataTransferFunction extends RichAsyncFunction<JSONObject, String> {

    static Logger logger = LoggerFactory.getLogger(AsyncMysqlDataTransferFunction.class);
    //数据库连接信息
    EnvProperties dbInfoMap;

    //线程池
    private transient ExecutorService executorService;

    public AsyncMysqlDataTransferFunction(EnvProperties dbInfoMap) {
        this.dbInfoMap = dbInfoMap;
    }

    @Override
    public void open(Configuration parameters) throws Exception {
        //初始化线程池
        executorService = new ThreadPoolExecutor(4 , 4, 20, TimeUnit.MINUTES, new LinkedBlockingDeque<>());
    }

    @Override
    public void close() throws Exception {
        executorService.shutdown();
    }

    @Override
    public void asyncInvoke(JSONObject value, ResultFuture<String> resultFuture) throws Exception {
        executorService.submit(() -> {
            try {
                String type = value.getString("type");
                JSONArray dataList = value.getJSONArray("data");
                JSONObject mysqlType = value.getJSONObject("mysqlType");
                JSONArray oldDataList = value.getJSONArray("old");
                String table = value.getString("table");
                Boolean isDdl = value.getBoolean("isDdl");
                JSONArray pkNames = value.getJSONArray("pkNames");
                JSONObject dataObj = dataList.getJSONObject(0);
                Set<String> pkNameSet = new HashSet<>();
                if(CollUtil.isNotEmpty(pkNames)){
                    pkNames.forEach(name -> pkNameSet.add(String.valueOf(name)));
                }

                String shardingRule = dataObj.getString("shardingRule");
                String concat = "";
                if (StrUtil.isNotBlank(shardingRule)){
                    Map<String,Object> map = JSON.parseObject(shardingRule, Map.class);
                    String strategy = map.get("strategy").toString();

                    Map<String,Object> strategyMap = JSON.parseObject(strategy, Map.class);
                    String str = MapUtil.getStr(strategyMap, "sharding-column");
                    int i = MapUtil.getInt(strategyMap, "sharding-count").intValue();

                    if (i > 1){
                        int tableNum = dataObj.getInteger(str).intValue() % i;
                        concat = table.concat("_").concat(String.valueOf(tableNum));
                    }else {
                        concat = table;
                    }

                }else {
                    String shardedTable = dbInfoMap.getSharded_table();
                    List<Map<String,Object>> list = JSONArray.parseObject(shardedTable, List.class);

                    List<Map<String, Object>> table1 = list.stream().filter(l -> {
                        return l.get("table").toString().equals(table);
                    }).collect(Collectors.toList());

                    if (ObjectUtils.isEmpty(table1)){
                        concat = table;
                    }else {
                        //分表数
                        String tbNum = table1.get(0).get("tb_num").toString();
                        //分表字段
                        Integer majorKey = dataObj.getInteger(table1.get(0).get("major_key").toString());
                        if (!tbNum.equals("1")){
                            int i1 = majorKey.intValue() % Integer.parseInt(tbNum);
                            concat = table.concat("_").concat(String.valueOf(i1));
                        }else {
                            concat = table;
                        }
                    }
                }

                String excueteSql = "";
                if(isDdl){
                    excueteSql = value.getString("sql");
                    if(StrUtil.isNotBlank(excueteSql)){
                        excueteSql = StrUtil.subBefore(excueteSql,"AFTER",true);
                    }
                    resultFuture.complete(Collections.singleton(excueteSql));
                    return;
                }

                if("INSERT".equals(type)){
                    excueteSql = tranferInsertSql(concat,dataObj,mysqlType);
                }
                if("UPDATE".equals(type)){
                    JSONObject oldDataObj = oldDataList.getJSONObject(0);
                    // excueteSql = tranferUpdateSql(table,dataObj,oldDataObj,mysqlType,pkNameSet);
                    excueteSql = tranferInsertSql(concat,dataObj,mysqlType);
                }

                if("DELETE".equals(type)){
                    excueteSql = transferDeleteSql(concat,dataObj,mysqlType,pkNameSet);
                }
                resultFuture.complete(Collections.singleton(excueteSql));

            }catch (Exception e){
                e.printStackTrace();
                resultFuture.complete(Collections.singleton("err"));
            }finally {

            }
        });
    }


    private static final String[] STR_SQL_TYPE = new String[]{"VARCHAR","CHAR","TINYBLOB","BLOB","MEDIUMBLOB","LONGBLOB","TINYTEXT","TEXT","MEDIUMTEXT","LONGTEXT","TIME","TIMESTAMP","JSON","json"};

    private static final String[] KEYWORD = new String[]{"limit"};
    private static String tranferInsertSql(String table, JSONObject dataObj, JSONObject mysqlType) {

        Set<String> columnSet = mysqlType.keySet();
        StringBuilder sb = new StringBuilder();
        for (String s : columnSet) {
            sb.append("`"+s+"`").append(",");
        }
        List<String> valueList = new ArrayList<>();
        List<String> updateList = new ArrayList<>();
        for (String col : columnSet) {
            String formatCol = "`"+col+"`";
            if(Arrays.asList(KEYWORD).contains(col)){
                valueList.add(getValueString(dataObj,col,mysqlType.getString(col)));
                updateList.add(formatCol.concat(" = VALUES(").concat(formatCol).concat(")"));
            }else {
                valueList.add(getValueString(dataObj,col,mysqlType.getString(col)));
                updateList.add(col.concat(" = VALUES(").concat(col).concat(")"));
            }
        }
        //String columnString = String.join(",",columnSet);
        sb.setLength(sb.length()-1);
        String columnString = sb.toString();
        String valueString = String.join(",",valueList);
        String updateString = String.join(",",updateList);

        return String.format("INSERT INTO %s (%s) values (%s) ON DUPLICATE KEY UPDATE %s;",table,columnString,valueString,updateString);
    }

    private String tranferUpdateSql(String table, JSONObject dataObj, JSONObject oldDataObj, JSONObject mysqlType,Set<String> pkNameSet) {

        Set<String> columnSet = mysqlType.keySet();
        List<String> setList = new ArrayList<>();
        List<String> whereList = new ArrayList<>();
        for (String col : columnSet) {
            String setString = col.concat(" = ").concat(getValueString(dataObj,col,mysqlType.getString(col)));
            setList.add(setString);
        }

        for (String pk : pkNameSet) {
            String whereString = pk.concat(" = ").concat(getValueString(oldDataObj,pk,mysqlType.getString(pk)));
            whereList.add(whereString);
        }
        String setString = String.join(",",setList);
        String whereString = String.join(" and ",whereList);

        return String.format("UPDATE %s SET %s WHERE %s",table,setString,whereString);
    }

    private String transferDeleteSql(String table, JSONObject dataObj, JSONObject mysqlType, Set<String> pkNameSet) {

        List<String> whereList = new ArrayList<>();
        for (String pk : pkNameSet) {
            String whereString = pk.concat(" = ").concat(getValueString(dataObj,pk,mysqlType.getString(pk)));
            whereList.add(whereString);
        }
        String whereString = String.join(" and ",whereList);
        return String.format("DELETE FROM %s WHERE %s",table,whereString);
    }

    /**
     * @author shezaixing
     * @date 2023/12/7 14:23
     * @description 判断拼接字符串时类型（是否需要加上引号）
     *
     */
    private static String getValueString(JSONObject dataObj,String columnKey,String mysqlType){

        if(null == dataObj.get(columnKey)){
            return "null";
        }

        //需要处理成字符串加引号的类型
        if(Arrays.asList(STR_SQL_TYPE).contains(mysqlType.toUpperCase())){
            return String.format("'%s'", dataObj.getString(columnKey).replace("\\","\\\\").replace("'", "\\'") );
        }

        //时间字段处理
        if("DATE".equalsIgnoreCase(mysqlType) || "DATETIME".equalsIgnoreCase(mysqlType)){
            SimpleDateFormat df = "DATETIME".equalsIgnoreCase(mysqlType) ? new SimpleDateFormat("yyyy-MM-dd HH:mm:ss") : new SimpleDateFormat("yyyy-MM-dd");
            return String.format("\"%s\"",df.format(dataObj.getDate(columnKey)));
        }

        return dataObj.getString(columnKey);
    }


    public static void main(String[] args) {

        int i = 1422;
        int i1 = i % 6 % 3;
        int i2 = i % 3;
        System.out.println(i1);
        System.out.println(i2);
//        JSONObject jsonObject = new JSONObject();
//
//        jsonObject.put("id",1);
//        jsonObject.put("name","Nana");
//        jsonObject.put("age",26);
//        jsonObject.put("salary",20000);
//        jsonObject.put("date1","2023-10-01");
//        jsonObject.put("date2","2023-10-02 11:11:00");
//
//        JSONObject mysqlType = new JSONObject();
//        mysqlType.put("id","int");
//        mysqlType.put("name","varchar");
//        mysqlType.put("age","bigint");
//        mysqlType.put("salary","double");
//        mysqlType.put("date1","date");
//        mysqlType.put("date2","datetime");
//        mysqlType.put("relation",null);
//
//        String table = "test";
//
//        String s= "ff8940af-c080-40cc-9d83-8c7dc8b86ed4";
//        System.out.println(s.length());
//
//        String s1 = "hello string sss";
//
//        String s2 = "'kaskljsl'";
//
//        System.out.println(StrUtil.subBefore(s1,"string",true));
//
//        System.out.println(tranferInsertSql(table,jsonObject,mysqlType));
//
//        System.out.println(s2.replaceAll("'","\\\\'"));
//
//        String[] ss = new String[]{"1","2","3" };
//        StringBuilder sb = new StringBuilder();
//        for (String s3 : ss)
//            {
//                sb.append("`"+s3+"?").append(",");
//        }
//        for (String s3 : ss) {
//            System.out.println(s3);
//        }
//        sb.setLength(sb.length()-1);
//        System.out.println(sb.toString());
//
//        String s5 = "交货地址：安徽霍邱供应站及指定地点\\\",\\\"bjsm\\\":null,\\\"jhType\\\":null,\\";
//        System.out.println(s5);
//        System.out.println(s5.replace("\\","\\\\").replace("'", "\\'"));
    }


}
